import argparse
import json
import os
import csv

from rdkit import Chem
from tqdm.auto import tqdm
from typing import Set

def contains_element(smile: str) -> bool:
    """
    Check if a SMILES string contains only specific atomic symbols.

    Args:
        smile (str): A SMILES string representing a molecule.

    Returns:
        bool: True if the molecule only contains the atomic symbols ['C', 'H', 'O', 'S', 'N', 'P', 'Cl'], False otherwise.
    """
    mol = Chem.MolFromSmiles(smile)
    atoms = [atom.GetSymbol() for atom in mol.GetAtoms()]
    unique_atoms = set(atoms)
    supply_element = ['C', 'H', 'O', 'S', 'N', 'P', 'Cl']
    valid_symbols = set(supply_element)
    return unique_atoms.issubset(valid_symbols)


def gen_smiles_main(save_name: str, num_mol: int, csv_file: str, min_atom_count: int = 39, max_atom_count: int = 120) -> None:
    """
    Main function to generate molecules and save them in a json file.

    Args:
        save_name (str): The path of the json file to save the molecules.
        num_mol (int): The number of molecules to be generated.
        csv_file (str): The path to the CSV file containing molecular data.
        min_atom_count (int, optional): The minimum number of atoms for a molecule. Defaults to 39.
        max_atom_count (int, optional): The maximum number of atoms for a molecule. Defaults to 120.
    """
    # If results file doesn't exist, initialize an empty dictionary.
    # Otherwise, load existing results.
    len_dict = 0
    smiles_dict = {}
    if os.path.isfile(save_name):
        with open(save_name, 'r') as f:
            smiles_dict = json.load(f)
        if len(smiles_dict) < num_mol:
            print(f"Only {len(smiles_dict)} molecules found, generating more...")
            len_dict = len(smiles_dict)
        else:
            print(f"Found {len(smiles_dict)} molecules, skipping generation...")
            return

    pbar = tqdm(total=num_mol - len_dict, desc="Generating molecules", leave=False)
    i = 0
    with open(csv_file, newline='') as csvfile:
        reader = csv.reader(csvfile, delimiter=',', quotechar='"')
        next(reader)  # Skip header
        for row in tqdm(reader):  # tqdm adds a progress bar
            try:
                cid = row[0]
                smile = row[1]
                atom_count = int(row[2])
                if smile not in smiles_dict.keys():
                    if max_atom_count >= atom_count >= min_atom_count and contains_element(smile):
                        smiles_dict[f"cid{cid}"] = smile
                        i += 1
                        pbar.update(1)
                if i >= num_mol:
                    break
            except ValueError:
                continue

    # Save the dictionary
    with open(save_name, 'w') as f:
        json.dump(smiles_dict, f)



from SiPMai.gen_data.smile_generation import gen_smiles_main
from SiPMai.gen_data.ray_generation import ray_gen_main
from SiPMai.gen_data.prepare_dataset import gen_index_main
import os
import json
from typing import Dict


def gen_all_data(smiles_file: str, num_mol: int, csv_file: str, min_atom: int, max_atom: int, num_cpus: int,
                 mol_save_dir: str,
                 resolution: int, blur_sigma: int, use_motion_blur: bool, use_gaussian_noise: bool,
                 gen_original_img: bool, gen_mol_drawing:bool, img_show: bool,
                 train_ratio: float, val_ratio: float, test_ratio: float, split_seed: int,
                 development: bool = False) -> None:
    """
    Generate all data required for the molecular images.

    Args:
        smiles_file (str): The path to the file containing SMILES strings.
        num_mol (int): The number of molecules to be generated.
        csv_file (str): The path to the CSV file containing molecular data.
        min_atom (int): The minimum number of atoms for a molecule.
        max_atom (int): The maximum number of atoms for a molecule.
        num_cpus (int): The number of CPUs to be used for data generation.
        mol_save_dir (str): The directory where the generated data should be saved.
        resolution (int): The resolution of the generated images.
        blur_sigma (int): The sigma value for the Gaussian blur applied to the images.
        use_motion_blur (bool): Whether to use motion blur in the images.
        use_gaussian_noise (bool): Whether to add Gaussian noise to the images.
        gen_original_img (bool): Whether to generate original images.
        img_show (bool): Whether to display the generated images.
        train_ratio (float): The ratio of data to be used for training.
        val_ratio (float): The ratio of data to be used for validation.
        test_ratio (float): The ratio of data to be used for testing.
        split_seed (int): The seed for the random splitting of data.
        development (bool, optional): Whether to run in development mode. Defaults to False.

    Raises:
        FileNotFoundError: If the smiles file does not exist.
    """
    if development:
        gen_smiles_main(smiles_file, num_mol, csv_file, min_atom, max_atom)
        print("smiles generated")

    if not os.path.exists(smiles_file):
        raise FileNotFoundError(f"smiles file {smiles_file} not found, please generate smiles first.")
    else:
        with open(smiles_file, 'r') as f:
            smiles_dict: Dict = json.load(f)
            num_mol_ = len(smiles_dict)
            print(f"Found {num_mol_} molecules!")
            if num_mol_ > num_mol:
                print(f"use the first {num_mol} molecules for later steps")
                smiles_dict = {k: smiles_dict[k] for k in list(smiles_dict)[:num_mol]}
    if num_cpus == 0:
        num_cpus = os.cpu_count()

    ray_gen_main(smiles_dict, mol_save_dir,
                 resolution, blur_sigma, use_motion_blur, use_gaussian_noise, gen_original_img, gen_mol_drawing,
                 num_cpus, img_show, development)
    gen_index_main(mol_save_dir, train_ratio, val_ratio, test_ratio, split_seed)


def main() -> None:
    """
    Main function that parses command line arguments and calls the gen_all function.
    """
    print("Start generating data with preset parameters (100k dataset with 39 <= num_atom <= 200)... ")
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--development", type=bool, default=False,
                        help="Whether to run in development mode. Development mode will run ray locally with serial output."
                             "Defaults to False.")
    # gen smiles args
    parser.add_argument("--smiles_file", type=str, default="pubchem_39_200_100k.json",
                        help="The path to the file containing SMILES strings.")
    parser.add_argument("--num_mol", type=int, default=100000, help="The number of molecules to be generated.")
    parser.add_argument("--csv_file", type=str, default="../CID-SMILES2.csv",
                        help="The path to the CSV file containing molecular data.")
    parser.add_argument("--min_atom", type=int, default=39, help="The minimum number of atoms for a molecule.")
    parser.add_argument("--max_atom", type=int, default=200, help="The maximum number of atoms for a molecule.")

    # gen ray args

    parser.add_argument("--mol_save_dir", type=str, default="pubchem_39_200_100k",
                        help="The directory where the generated data should be saved.")

    parser.add_argument("--resolution", type=int, default=256, help="The resolution of the generated images.")
    parser.add_argument("--blur_sigma", type=int, default=16,
                        help="The sigma value for the Gaussian blur applied to the images.")
    parser.add_argument("--use_motion_blur", type=bool, default=False, help="Whether to use motion blur in the images.")
    parser.add_argument("--use_gaussian_noise", type=bool, default=False,
                        help="Whether to add Gaussian noise to the images.")
    parser.add_argument("--gen_original_img", type=bool, default=True, help="Whether to generate original images.")
    parser.add_argument("--gen_mol_drawing", type=bool, default=True, help="Whether to generate mol images for eye inspection.")
    parser.add_argument("--num_cpus", type=int, default=0, help="The number of CPUs to be used for data generation.")
    parser.add_argument("--show", type=bool, default=False, help="Whether to display the generated images.")

    # data indices args
    parser.add_argument("--split_seed", type=int, default=1, help="The seed for the random splitting of data.")
    parser.add_argument("--train_ratio", type=float, default=0.8, help="The ratio of data to be used for training.")
    parser.add_argument("--val_ratio", type=float, default=0.1, help="The ratio of data to be used for validation.")
    parser.add_argument("--test_ratio", type=float, default=0.1, help="The ratio of data to be used for testing.")
    args = parser.parse_args()

    from pkg_resources import resource_filename
    args.smiles_file = resource_filename('SiPMai', f'smiles/{args.smiles_file}')
    cmd_dir = os.getcwd()
    args.mol_save_dir = os.path.join(cmd_dir, args.mol_save_dir)
    print("mol_save_dir set to command execution dir: ", args.mol_save_dir)
    gen_all_data(smiles_file=args.smiles_file, num_mol=args.num_mol, csv_file=args.csv_file, min_atom=args.min_atom,
                 max_atom=args.max_atom, num_cpus=args.num_cpus, mol_save_dir=args.mol_save_dir,
                 resolution=args.resolution, blur_sigma=args.blur_sigma, use_motion_blur=args.use_motion_blur,
                 use_gaussian_noise=args.use_gaussian_noise, gen_original_img=args.gen_original_img, gen_mol_drawing=args.gen_mol_drawing,
                 img_show=False,
                 train_ratio=args.train_ratio, val_ratio=args.val_ratio, test_ratio=args.test_ratio,
                 split_seed=args.split_seed,
                 development=args.development)


if __name__ == "__main__":
    main()



import os
import json
import numpy as np
from sklearn.model_selection import train_test_split
import warnings
import argparse
from tqdm import tqdm

def generate_index_dict(ROOT_DIR):
    IMG_DIR, INFO_DIR, JSON_DIR = os.path.join(ROOT_DIR, 'img'), os.path.join(ROOT_DIR, 'info'), os.path.join(ROOT_DIR, 'json')
    # load all file names in the directory
    img_names = [name for name in os.listdir(IMG_DIR)
                 if os.path.isfile(os.path.join(IMG_DIR, name)) and "orig_img" not in name]
    info_names = [name for name in os.listdir(INFO_DIR)
                  if os.path.isfile(os.path.join(INFO_DIR, name)) and name.endswith("npz")]
    json_names = [name for name in os.listdir(JSON_DIR)
                  if os.path.isfile(os.path.join(JSON_DIR, name))]

    # check if the number of files are the same
    if not len(img_names) == len(info_names) == len(json_names):
        warnings.warn("The number of files in the directories are not the same! Rerun mol generation")

    molecule_dict = {}
    for img, info, json_file in tqdm(zip(sorted(img_names), sorted(info_names), sorted(json_names)),
                                     desc="Generating index dict"):
        # extract molecule id from the file name
        mol_id = "_".join(img.split('_')[:-1])
        if os.path.exists(os.path.join(IMG_DIR, img)) \
                and os.path.exists(os.path.join(INFO_DIR, info))\
                and os.path.exists(os.path.join(JSON_DIR, json_file)):
            molecule_dict[mol_id] = {
                "img_path": os.path.join('img', img),  # under the ROOT_DIR to avoid relative path problem
                "info_path": os.path.join('info', info),
                "json_path": os.path.join('json', json_file)
            }

    return molecule_dict

def gen_index_main(ROOT_DIR, train_ratio, val_ratio, test_ratio, seed):
    mol_dict = generate_index_dict(ROOT_DIR)
    # List all the molecules
    # Set the splitting ratios
    assert abs(train_ratio + val_ratio + test_ratio - 1) < 1e-5, "The sum of ratios should be 1"
    # Split the molecules
    train, test = train_test_split(list(mol_dict.keys()), test_size=1 - train_ratio, random_state=seed)
    val, test = train_test_split(test, test_size=test_ratio / (test_ratio + val_ratio), random_state=seed)

    # Generate dictionaries
    train_dict = {mol: mol_dict[mol] for mol in train}
    val_dict = {mol: mol_dict[mol] for mol in val}
    test_dict = {mol: mol_dict[mol] for mol in test}

    # save dictionaries as json files
    with open(os.path.join(ROOT_DIR, 'train_set_index.json'), 'w') as fp:
        json.dump(train_dict, fp)

    with open(os.path.join(ROOT_DIR, 'val_set_index.json'), 'w') as fp:
        json.dump(val_dict, fp)

    with open(os.path.join(ROOT_DIR, 'test_set_index.json'), 'w') as fp:
        json.dump(test_dict, fp)
    print("dataset indices generated!")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Example of argparse usage.')
    parser.add_argument("--ROOT_DIR", type=str, default="../data")
    parser.add_argument("--seed", type=int, default=1)
    parser.add_argument("--train_ratio", type=float, default=0.8)
    parser.add_argument("--val_ratio", type=float, default=0.1)
    parser.add_argument("--test_ratio", type=float, default=0.1)
    args = parser.parse_args()
    gen_index_main(args.ROOT_DIR, args.train_ratio, args.val_ratio, args.test_ratio, args.seed)



from matplotlib.path import Path
import numpy as np
import math
from numba import jit, njit
from numba.typed import List


def cal_atom_projection(atom_position, mesh, z_min, get_coincidence=False):
    """
    :param atom_position: n_atom (x, y, z, r, ) of the atom
    :param mesh: [H, W, 2] mesh grid of x, y axes
    :param z_min: the minimum z coordinate of the atom
    :return: height [n_atom, H, W], atom mask [n_atom, H, W], bool [H, W], the 1st dim order is the order of atom_position
    """
    def ball_3d(x0, y0, z0, r, x, y):
        # Compute squared distance from the sphere center to the point in the xy-plane
        sq_dist_xy = (x - x0) ** 2 + (y - y0) ** 2
        sq_dist_xy[sq_dist_xy > r ** 2] = r ** 2
        # Calculate z-coordinate (height) of the point on the sphere surface
        z = z0 + np.sqrt(r ** 2 - sq_dist_xy)

        return z

    def compute_height_matrix(xyzr, mesh):
        # xyzr shape: [n, 4] (x0, y0, z0, r for each sphere)
        # mesh shape: [H, W, 2] (x, y coordinate for each point in the mesh)


        n_spheres = xyzr.shape[0]
        H, W = mesh.shape[0], mesh.shape[1]

        height_matrix = np.zeros((n_spheres, H, W))

        for i in range(n_spheres):
            x0, y0, z0, r = xyzr[i]

            # Compute the height for the sphere and store it in the height matrix
            height_matrix[i] = ball_3d(x0, y0, z0, r, mesh[:, :, 0], mesh[:, :, 1])
        return height_matrix

    def check_coincidence(height_matrix):
        # Compute maximum and minimum height for each point in the mesh across all spheres
        pos_mask, neg_mask = height_matrix.copy(), height_matrix.copy()
        pos_mask[pos_mask > 0] = 1
        pos_mask = pos_mask.sum(axis=0)

        neg_mask = neg_mask.sum(axis=0)
        neg_mask[neg_mask > 0] = 1

        coincide_matrix = (pos_mask - neg_mask).astype(bool)
        return coincide_matrix


    xyzr, atom_class = atom_position[:, :4], atom_position[:, 4]
    height_all_atom = compute_height_matrix(xyzr, mesh)
    height = np.max(height_all_atom, axis=0)

    is_coincide = check_coincidence(height_all_atom) if get_coincidence else None
    atom_mask = height_all_atom > 0
    return height + z_min, atom_mask, is_coincide


def bond_location_cal(bond_list, mesh, points_bond_bool, resolution):
    @jit(nopython=True)
    def cal_bond_atom_loc(point1_x, point1_y, point2_x, point2_y, a1_radius, a2_radius):
        bond_height = math.sqrt(math.pow((point2_x - point1_x), 2) + math.pow((point2_y - point2_x), 2)) / 3
        bond_x = (point1_x + point2_x) / 2
        bond_y = (point1_y + point2_y) / 2
        bond_width = (a1_radius + a2_radius) / 6
        dx = point2_x - point1_x
        dy = point2_y - point1_y
        angle = math.atan2(dy, dx)
        return bond_y, bond_x, angle, bond_height, bond_width

    for b in bond_list:
        point1_x, point1_y = b[1], b[2]
        point2_x, point2_y = b[4], b[5]
        a1_radius, a2_radius = b[3], b[6]
        bond_y, bond_x, angle, bond_height, bond_width = cal_bond_atom_loc(point1_x, point1_y, point2_x, point2_y,
                                                                           a1_radius, a2_radius)
        # Determine the bond area
        bond_loc = rect_loc(bond_y=bond_y, bond_x=bond_x, bond_angle=angle, bond_height=bond_height,
                            bond_width=bond_width)
        bond_area = Path(bond_loc)
        for r in range(resolution):
            for t in range(resolution):
                x_1 = mesh[r, t, 0]
                y_1 = mesh[r, t, 1]
                points_bond_bool[b[0], r, t] = bond_area.contains_point((x_1, y_1))
    return points_bond_bool


@jit(nopython=True)
def rect_loc(bond_y, bond_x, bond_angle, bond_height, bond_width):
    """
    @param bond_y: The center of the bond y value
    @param bond_x: The center of the bond x value
    @param bond_angle: The slope of the line between the atom and the origin
    @param bond_height: Bond length
    @param bond_width: Bond width
    @return: Bond area
    """
    xo = np.cos(bond_angle)
    yo = np.sin(bond_angle)
    y1 = bond_y + bond_height / 2 * yo
    x1 = bond_x + bond_height / 2 * xo
    y2 = bond_y - bond_height / 2 * yo
    x2 = bond_x - bond_height / 2 * xo

    return np.array(
        [
            [x1 + bond_width / 2 * yo, y1 - bond_width / 2 * xo],
            [x2 + bond_width / 2 * yo, y2 - bond_width / 2 * xo],
            [x2 - bond_width / 2 * yo, y2 + bond_width / 2 * xo],
            [x1 - bond_width / 2 * yo, y1 + bond_width / 2 * xo],
        ]
    )



import cv2
import numpy as np


def uniform_noise(img):
    mean, sigma = 10, 100
    a = 2 * mean - np.sqrt(12 * sigma)  # a = -14.64
    b = 2 * mean + np.sqrt(12 * sigma)  # b = 54.64
    noiseUniform = np.random.uniform(a, b, img.shape)
    imgUniformNoise = img + noiseUniform
    imgUniformNoise = np.uint8(cv2.normalize(imgUniformNoise, None, 0, 255, cv2.NORM_MINMAX))  # 归一化为 [0,255]
    imgUniformNoise = np.array(imgUniformNoise, dtype=np.uint8)
    return imgUniformNoise


def motion_blur(image, degree=12, angle=45):
    image = np.array(image)
    # 这里生成任意角度的运动模糊kernel的矩阵， degree越大，模糊程度越高
    M = cv2.getRotationMatrix2D((degree / 2, degree / 2), angle, 1)
    motion_blur_kernel = np.diag(np.ones(degree))
    motion_blur_kernel = cv2.warpAffine(motion_blur_kernel, M, (degree, degree))

    motion_blur_kernel = motion_blur_kernel / degree
    blurred = cv2.filter2D(image, -1, motion_blur_kernel)

    # convert to uint8
    cv2.normalize(blurred, blurred, 0, 255, cv2.NORM_MINMAX)
    blurred = np.array(blurred, dtype=np.uint8)
    return blurred


def compress_binary_arr(arr):
    """Compress a binary tensor to int8 array"""
    flattened = arr.reshape(-1)
    int8_representation = np.packbits(flattened)
    return int8_representation


def decompress_binary_arr(npz_filename):
    # Loading compressed data and decompressing it
    data = np.load(npz_filename)
    arr_atom = data['arr_atom']
    arr_bound = data['arr_bond']
    adj_matrix = data['adj_matrix']
    molecule_points_height = data['molecule_points_height']

    arr_atom_shape = data['arr_atom_shape']
    arr_bond_shape = data['arr_bond_shape']

    # """Decompress an int8 array back to a binary tensor"""
    arr_atom = np.unpackbits(arr_atom)
    arr_atom = arr_atom.reshape(arr_atom_shape)

    arr_bound = np.unpackbits(arr_bound)
    arr_bound = arr_bound.reshape(arr_bond_shape)

    return arr_atom, arr_bound, adj_matrix, molecule_points_height



import warnings
import os
import json
from tqdm import tqdm
from rdkit import Chem
from rdkit.Chem import Draw
from SiPMai.gen_data.util_fn import uniform_noise, motion_blur, compress_binary_arr, decompress_binary_arr
from matplotlib import pyplot as plt
from PIL import Image
import numpy as np
from scipy.ndimage import gaussian_filter
from SiPMai.gen_data.compute_img import cal_atom_projection, bond_location_cal
if not "ENABLE_MAYAVI" in globals():
    try:
        from mayavi import mlab

        ENABLE_MAYAVI = True
    except ImportError as e:
        ENABLE_MAYAVI = False
        warnings.warn("mayavi not found, this will not hinder data generation. "
                      "If img_show is needed, please install mayavi")

def points_height_matrix(smi: str, molecule_name: int, resolution: int, info_dir, json_dir, img_dir,
                         blur_sigma, use_motion_blur, use_gaussian_noise,
                         show=False):
    """
    generate 5 files for each molecule:
    1. img_dir/img.png:                 the image of the molecule with gaussian noise (for simulating electron cloud)
    2. img_dir/orig_img.png:            the image of the molecule without noise, calculated from the 2D coordinates
    3. info_dir/points_info.npz:   -- atom instance binary array, shape: (num_atom, resolution, resolution)
                                         -- bond instance binary array, shape: (num_bond, resolution, resolution)
                                         -- adjacency matrix, shape: (num_atom, num_atom)
                                         -- point height array, shape: (resolution, resolution)
    4. info_dir/mol_drawing.png   a molecule image (only for human inspection)
    5. json_dir/.json                   json file containing the SMILES string and other useful information
    """
    img_name = os.path.join(img_dir, str(molecule_name) + '_img.png')
    orig_img_name = os.path.join(img_dir, str(molecule_name) + '_orig_img.png')
    points_info_name = os.path.join(info_dir, str(molecule_name) + '_points_info.npz')
    drawing_name = os.path.join(info_dir, str(molecule_name) + '_mol_drawing.png')
    json_name = os.path.join(json_dir, str(molecule_name) + '.json')

    # check done
    if os.path.exists(points_info_name) and os.path.exists(img_name) and os.path.exists(
            orig_img_name) and os.path.exists(
            drawing_name) and os.path.exists(json_name):
        print(f"molecule {molecule_name} exists, skip")
        return True

    #         2D molecules were converted into 3D structures,
    #         and mechanical functions were used for Angle correction and optimization
    smile_dict = {}
    mol_2D = Chem.MolFromSmiles(smi)
    smile_dict['smile'] = smi
    mol_2D_H = Chem.AddHs(mol_2D)
    atom_num = mol_2D_H.GetNumAtoms()
    bond_num = mol_2D_H.GetNumBonds()

    moleblock_2D = Chem.MolToMolBlock(mol_2D_H)  # Returns the two-dimensional coordinates of the atoms in the molecule
    if show:
        mol_3D = Chem.MolFromSmiles(smi)
        Draw.ShowMol(mol_3D, size=(550, 550), kekulize=False)
        Draw.ShowMol(mol_2D_H, size=(550, 550), kekulize=False)
    mol_2D_H = Chem.MolFromMolBlock(moleblock_2D, removeHs=False)
    conf = mol_2D_H.GetConformer()

    atom_position = []
    atom_radius = {'C': 0.77, 'H': 0.37, 'O': 0.73, 'S': 1.02, 'N': 0.75, 'P': 1.06, 'Cl': 0.99}
    atom_class = {'0.77': 1, '0.37': 2, '0.73': 3, '1.02': 4, '0.75': 5, '1.06': 6, '0.99': 7}
    for s in range(atom_num):
        x = list(conf.GetAtomPosition(s))
        mol_2D_H.GetAtomWithIdx(s).SetProp('molAtomMapNumber', str(mol_2D_H.GetAtomWithIdx(s).GetIdx()))  # 对原子进行索引
        atom_symbol = mol_2D_H.GetAtoms()[s].GetSymbol()
        if atom_symbol in atom_radius.keys():
            x.append(atom_radius[atom_symbol])
        else:
            raise NotImplementedError("only support C, H, O, S, N, P, Cl")
        x.append(mol_2D_H.GetAtomWithIdx(s).GetIdx())
        atom_position.append(x)
    Draw.MolToFile(mol_2D_H, drawing_name, size=(600, 600))
    # 在三维坐标系绘制得到的原子坐标，可通过原子坐标调整三维坐标轴大小
    if show and ENABLE_MAYAVI:
        plt.figure()  # 得到画面
        ax1 = plt.axes(projection='3d')
        ax1.set_xlim(0, 10)  # X轴，横向向右方向
        ax1.set_ylim(10, 0)  # Y轴,左向与X,Z轴互为垂直
        ax1.set_zlim(0, 10)  # 竖向为Z轴
        # color1 = ['r', 'g', 'b', 'k', 'm']
        # marker1 = ['o', 'v', '1', 's', 'H']
        for x in atom_position:
            ax1.scatter(x[0], x[1], x[2], c='r', marker='o', linewidths=4)
        plt.show()

    points = np.array(atom_position)

    # 计算基平面上的投影的x，y的坐标范围，确定生成的坐标点阵的区域
    x, y, z, r = points[:, 0], points[:, 1], points[:, 2], points[:, 3]
    r_max = max(r)
    x_max, x_min, y_max, y_min = int(max(x) + r_max) + 1, int(min(x) - r_max) - 1, int(max(y) + r_max) + 1, int(
        min(y) - r_max) - 1
    z_min = int(min(z) - r_max) - 1  # 方便后续对基平面进行平移操作

    # Draw a scale model of a 3D molecule (stacked ball model)
    if show and ENABLE_MAYAVI:
        mlab.points3d(x, y, z, r * 2, scale_factor=1, resolution=30, mode="sphere")
        mlab.outline()
        mlab.axes()
        mlab.show()


    # Initializes the bool matrix that stores the atoms
    points_bond_bool = np.zeros([bond_num, resolution, resolution])

    # create coordinate mesh
    x_axes = np.linspace(x_min, x_max, resolution)
    y_axes = np.linspace(y_min, y_max, resolution)
    X, Y = np.meshgrid(x_axes, y_axes)
    coordinate_points = np.array([X.ravel(), Y.ravel()])
    coordinate_points_array = coordinate_points.T
    coordinate_points_array = np.array(coordinate_points_array).reshape((resolution, resolution, 2))

    '''
    Start to calculate the height of each atom in the molecule from the base plane. At this time, 
    the base plane is set as the xoy plane, and the virtual projection is made from top to bottom through 
    the point light source cluster, that is, the height of the highest atom contacted is taken as the height recorded
    '''
    points_initial, points_bool, points_class = cal_atom_projection(np.array(atom_position), coordinate_points_array,
                                                                    resolution,  atom_class, z_min)
    points_class = np.where(points_class == 0, 255, points_class)
    # attention! 图像坐标系的原点通常是左上角，而数组的索引是从左到右、从上到下的。因此，在将 NumPy 数组转换为 Pillow 图像对象时，需要将数组进行翻转，使其与图像坐标系保持一致。可以使用以下代码进行翻转：
    im = Image.fromarray(points_class[::-1, :])
    im = im.convert('L')
    im.save(orig_img_name)

    # get blurred image
    points_initial = gaussian_filter(points_initial, sigma=blur_sigma)
    if use_motion_blur:
        points_initial = motion_blur(points_initial)
    if use_gaussian_noise:
        points_initial = uniform_noise(points_initial)
    plt.figure(figsize=(resolution, resolution), dpi=1)
    plt.imshow(points_initial, cmap='gray', origin='lower')
    plt.axis('off')
    fig = plt.gcf()
    plt.gca().xaxis.set_major_locator(plt.NullLocator())
    plt.gca().yaxis.set_major_locator(plt.NullLocator())
    plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
    plt.margins(0, 0)
    fig.savefig(img_name, format='png')
    plt.close()

    # todo: calculate per pixel label of each bond in the image
    bond_list = [[] for i in range(bond_num)]
    molecule_adjacent_matrix = np.zeros((atom_num, atom_num), dtype=np.bool_)
    bond_record_dic = {}
    bonds = mol_2D_H.GetBonds()  # To traverse the key
    for bond_i in range(bond_num):
        bond_list[bond_i].append(bonds[bond_i].GetIdx())
        atom_begin_index = bonds[bond_i].GetBeginAtomIdx()
        atom_end_index = bonds[bond_i].GetEndAtomIdx()
        bond_record_dic[f'bond-{bond_i}'] = str(atom_begin_index) + str(atom_end_index)
        molecule_adjacent_matrix[atom_begin_index][atom_end_index] = 1
        molecule_adjacent_matrix[atom_end_index][atom_begin_index] = 1
        for t in atom_position:
            if t[4] == atom_begin_index or t[4] == atom_end_index:
                bond_list[bond_i].append(t[0])
                bond_list[bond_i].append(t[1])
                bond_list[bond_i].append(t[3])

    points_bond_bool = bond_location_cal(bond_list, coordinate_points_array, points_bond_bool, resolution)

    # save np.array

    arr_atom = np.array(points_bool, dtype=bool)
    arr_bond = np.array(points_bond_bool, dtype=bool)

    arr_atom_compressed = compress_binary_arr(arr_atom)
    arr_bond_compressed = compress_binary_arr(arr_bond)
    adj_matrix_compressed = compress_binary_arr(molecule_adjacent_matrix)

    np.savez_compressed(points_info_name,
                        arr_atom=arr_atom_compressed, arr_bond=arr_bond_compressed, adj_matrix=adj_matrix_compressed,
                        arr_atom_shape=arr_atom.shape, arr_bond_shape=arr_bond.shape,
                        adj_shape=molecule_adjacent_matrix.shape,
                        molecule_points_height=points_initial)

    # turn on debug mode to check compression
    arr_atom_, arr_bond_, adj, points_initial_ = decompress_binary_arr(points_info_name)
    assert np.array_equal(arr_atom, arr_atom_)
    assert np.array_equal(arr_bond, arr_bond_)
    assert np.array_equal(adj, molecule_adjacent_matrix)
    assert np.array_equal(points_initial, points_initial_)

    json_information = [bond_record_dic, smile_dict]
    with open(json_name, 'w', encoding='utf-8') as f:
        json.dump(json_information, f)
    return True


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--smiles_file", type=str, default="../data/pubchem_smiles.json")
    parser.add_argument("--save_dir", type=str, default="../data")
    parser.add_argument("--resolution", type=int, default=256)
    parser.add_argument("--blur_sigma", type=int, default=24)
    parser.add_argument("--use_motion_blur", type=bool, default=False)
    parser.add_argument("--use_gaussian_noise", type=bool, default=False)
    parser.add_argument("--show", type=bool, default=False)
    args = parser.parse_args()

    # start generation
    with open(args.smiles_file, 'r') as f:
        mol_dict = json.load(f)

    # prepare folders
    info_dir = os.path.join(args.save_dir, "info")
    json_dir = os.path.join(args.save_dir, "json")
    img_dir = os.path.join(args.save_dir, "img")
    os.makedirs(info_dir, exist_ok=True)
    os.makedirs(json_dir, exist_ok=True)
    os.makedirs(img_dir, exist_ok=True)

    pbar = tqdm(total=len(mol_dict), desc="Generating molecules")
    for mol, smiles in mol_dict.items():
        task_id = points_height_matrix(smiles, mol, args.resolution, info_dir, json_dir, img_dir,
                                       args.blur_sigma, args.use_motion_blur, args.use_gaussian_noise, args.show)
        pbar.update()



import os
import json

import ray
from tqdm import tqdm
import time
from rdkit import Chem
from rdkit.Chem import Draw
from SiPMai.gen_data.util_fn import uniform_noise, motion_blur, compress_binary_arr, decompress_binary_arr
from matplotlib import pyplot as plt
from PIL import Image
import numpy as np
from SiPMai.gen_data.compute_img import cal_atom_projection, bond_location_cal
from scipy.ndimage import gaussian_filter

ENABLE_MAYAVI = False


@ray.remote
def points_height_matrix(smi: str, molecule_name: int, resolution: int, info_dir, json_dir, img_dir,
                         blur_sigma, use_motion_blur, use_gaussian_noise,
                         gen_original_img, gen_mol_drawing,
                         show=False):
    """
    generate 5 files for each molecule:
    1. img_dir/img.png:                 the image of the molecule with gaussian noise (for simulating electron cloud)
    2. img_dir/orig_img.png:            binary image of the molecule (only for human inspection)
    3. info_dir/points_info.npz:   -- atom instance binary array, shape: (num_atom, resolution, resolution)
                                         -- bond instance binary array, shape: (num_bond, resolution, resolution)
                                         -- adjacency matrix, shape: (num_atom, num_atom)
                                         -- point height array, shape: (resolution, resolution)
    4. info_dir/mol_drawing.png   a molecule image (only for human inspection)
    5. json_dir/.json                   json file containing the SMILES string and other useful information
    """
    img_name = os.path.join(img_dir, str(molecule_name) + '_img.png')
    orig_img_name = os.path.join(img_dir, str(molecule_name) + '_orig_img.png')
    points_info_name = os.path.join(info_dir, str(molecule_name) + '_points_info.npz')
    drawing_name = os.path.join(info_dir, str(molecule_name) + '_mol_drawing.png')
    json_name = os.path.join(json_dir, str(molecule_name) + '.json')

    if show:
        raise NotImplementedError("show is not implemented for ray version")

    # check done
    check_list = [img_name, points_info_name, json_name]
    if gen_original_img:
        check_list.append(orig_img_name)
    if gen_mol_drawing:
        check_list.append(drawing_name)
    done = np.array([os.path.exists(file) for file in check_list])
    if done.all():
        print("molecule already exists, skipping...")
        return True

    #         2D molecules were converted into 3D structures,
    #         and mechanical functions were used for Angle correction and optimization
    mol_2D = Chem.MolFromSmiles(smi)
    mol_2D_H = Chem.AddHs(mol_2D)
    atom_num = mol_2D_H.GetNumAtoms()
    bond_num = mol_2D_H.GetNumBonds()

    moleblock_2D = Chem.MolToMolBlock(mol_2D_H)  # Returns the two-dimensional coordinates of the atoms in the molecule
    if show:
        mol_3D = Chem.MolFromSmiles(smi)
        Draw.ShowMol(mol_3D, size=(550, 550), kekulize=False)
        Draw.ShowMol(mol_2D_H, size=(550, 550), kekulize=False)
    mol_2D_H = Chem.MolFromMolBlock(moleblock_2D, removeHs=False)
    conf = mol_2D_H.GetConformer()

    atom_position = np.zeros((atom_num, 5), dtype=np.float32)
    atom_radius = {'C': 0.77, 'H': 0.37, 'O': 0.73, 'S': 1.02, 'N': 0.75, 'P': 1.06, 'Cl': 0.99}
    for s in range(atom_num):
        mol_2D_H.GetAtomWithIdx(s).SetProp('molAtomMapNumber', str(mol_2D_H.GetAtomWithIdx(s).GetIdx()))
        atom_symbol = mol_2D_H.GetAtoms()[s].GetSymbol()
        # get x, y, z, r
        if atom_symbol in atom_radius.keys():
            atom_position[s, 3] = atom_radius[atom_symbol]
        else:
            raise NotImplementedError("only support C, H, O, S, N, P, Cl")
        atom_position[s, :3] = list(conf.GetAtomPosition(s))
        atom_position[s, 4] = mol_2D_H.GetAtomWithIdx(s).GetIdx()  # atom index

    # save drawing
    if gen_mol_drawing:
        Draw.MolToFile(mol_2D_H, drawing_name, size=(600, 600))

    # get image boundaries
    x, y, z, r = atom_position[:, 0], atom_position[:, 1], atom_position[:, 2], atom_position[:, 3]
    r_max = r.max()
    x_max, x_min, y_max, y_min = int(x.max() + r_max) + 1, int(x.min() - r_max) - 1, int(y.max() + r_max) + 1, int(
        y.min() - r_max) - 1
    z_min = int(z.min() - r_max) - 1  # doesn't really matter

    # get x, y mesh
    points_bond_bool = np.zeros([bond_num, resolution, resolution], dtype=bool)
    x_axes = np.linspace(x_min, x_max, resolution)
    y_axes = np.linspace(y_min, y_max, resolution)
    X, Y = np.meshgrid(x_axes, y_axes)
    coordinate_points = np.array([X.ravel(), Y.ravel()])
    mesh = coordinate_points.T
    mesh = np.array(mesh).reshape((resolution, resolution, 2))

    '''
    Start to calculate the height of each atom in the molecule from the base plane. At this time, 
    the base plane is set as the xoy plane, and the virtual projection is made from top to bottom through 
    the point light source cluster, that is, the height of the highest atom contacted is taken as the height recorded
    '''
    height_mesh, atom_mask, _ = cal_atom_projection(atom_position, mesh, z_min, get_coincidence=False)
    # if is_coincide.any():
    #     raise ValueError("atom coincide")
    if gen_original_img:
        orig_arr = atom_mask.sum(axis=0)
        orig_arr[orig_arr > 0] = 255
        im = Image.fromarray(orig_arr.astype(np.uint8))
        im = im.convert('L')
        im.save(orig_img_name)

    # get blurred image
    height_mesh = gaussian_filter(height_mesh, sigma=blur_sigma)
    if use_motion_blur:
        height_mesh = motion_blur(height_mesh)
    if use_gaussian_noise:
        height_mesh = uniform_noise(height_mesh)
    plt.figure(figsize=(resolution, resolution), dpi=1)
    plt.imshow(height_mesh, cmap='gray', origin='lower')
    plt.axis('off')
    fig = plt.gcf()
    plt.gca().xaxis.set_major_locator(plt.NullLocator())
    plt.gca().yaxis.set_major_locator(plt.NullLocator())
    plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
    plt.margins(0, 0)
    fig.savefig(img_name, format='png')
    plt.close()

    # get adjacent matrix
    bonds = mol_2D_H.GetBonds()  # To traverse the key
    molecule_adjacent_matrix = np.zeros((atom_num, atom_num), dtype=np.bool_)
    bond_record_dic = {}
    bond_list = [[bond.GetIdx()] for bond in bonds]

    for bond_i, bond in enumerate(bonds):
        atom_begin_index = bond.GetBeginAtomIdx()
        atom_end_index = bond.GetEndAtomIdx()
        bond_key = f"{atom_begin_index}-{atom_end_index}"
        bond_record_dic[bond_key] = bond_i
        molecule_adjacent_matrix[atom_begin_index, atom_end_index] = 1
        molecule_adjacent_matrix[atom_end_index, atom_begin_index] = 1

        matching_positions = atom_position[
            (atom_position[:, 4] == atom_begin_index) | (atom_position[:, 4] == atom_end_index)]
        bond_list[bond_i].extend(matching_positions[:, [0, 1, 3]].ravel().tolist())

    # get bond mask in the image
    points_bond_bool = bond_location_cal(bond_list, mesh, points_bond_bool, resolution)

    # save np.array

    arr_atom = np.array(atom_mask, dtype=bool)
    arr_bond = np.array(points_bond_bool, dtype=bool)

    arr_atom_compressed = compress_binary_arr(arr_atom)
    arr_bond_compressed = compress_binary_arr(arr_bond)

    np.savez_compressed(points_info_name,
                        arr_atom=arr_atom_compressed, arr_bond=arr_bond_compressed, adj_matrix=molecule_adjacent_matrix,
                        molecule_points_height=height_mesh)
    try:
        with open(json_name, 'w', encoding='utf-8') as f:
            json.dump({"bond_dict": bond_record_dic, "smiles": smi}, f)
    except Exception as e:
        # to avoid the error of json.dump, remove the file if an exception occurs
        os.remove(json_name)
        raise ValueError("something wrong with json.dump. Json file removed to avoid saving broken files")
    return True


def ray_gen_main(mol_dict, save_dir, resolution, blur_sigma, use_motion_blur, use_gaussian_noise, gen_original_img=False, gen_mol_drawing=False,
                 num_cpu=None, show=False, debug_mode=False):
    ray.init(num_cpus=num_cpu, local_mode=debug_mode)

    # prepare folders
    info_dir = os.path.join(save_dir, "info")
    json_dir = os.path.join(save_dir, "json")
    img_dir = os.path.join(save_dir, "img")
    os.makedirs(info_dir, exist_ok=True)
    os.makedirs(json_dir, exist_ok=True)
    os.makedirs(img_dir, exist_ok=True)

    tasks = {}
    for mol, smiles in mol_dict.items():
        task_id = points_height_matrix.remote(smiles, mol, resolution, info_dir, json_dir, img_dir,
                                              blur_sigma, use_motion_blur, use_gaussian_noise,
                                              gen_original_img, gen_mol_drawing, show)
        tasks[task_id] = mol

    pbar = tqdm(total=len(tasks), desc="Processing tasks")
    while len(tasks):
        done_ids, _ = ray.wait(list(tasks.keys()), num_returns=1)
        for ready_id in done_ids:
            try:
                result = ray.get(ready_id)
                molecule_name = tasks.pop(ready_id)
                pbar.set_description(molecule_name)
                pbar.update()
            except ray.exceptions.RayTaskError as ex:
                print(f"Task failed for molecule {tasks[ready_id]} with error: {ex}")
        time.sleep(0.5)  # To avoid busy waiting

    pbar.close()
    ray.shutdown()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--smiles_file", type=str, default="../smiles/pubchem_39_200_100k.json")
    parser.add_argument("--save_dir", type=str, default="../data")
    parser.add_argument("--resolution", type=int, default=256)
    parser.add_argument("--blur_sigma", type=int, default=24)
    parser.add_argument("--use_motion_blur", type=bool, default=False)
    parser.add_argument("--use_gaussian_noise", type=bool, default=False)
    parser.add_argument("--show", type=bool, default=False)
    args = parser.parse_args()
    with open(args.smiles_file, 'r') as f:
        mol_dict = json.load(f)
    # start generation
    ray_gen_main(mol_dict, args.save_dir,
                 args.resolution, args.blur_sigma, args.use_motion_blur, args.use_gaussian_noise, args.show, num_cpu=2)



import os
from setuptools import find_packages, setup


def get_version() -> str:
    init = open(os.path.join("SiPMai", "__init__.py"), "r").read().split()
    return init[init.index("__version__") + 2][1:-1]


def get_install_requires():
    return [
        "numpy>=1.16,<1.24",
        "torch>=1.4.0",
        "packaging",
        "tqdm",
        "scikit-learn",
        "matplotlib",
        "scipy",
        "pandas",
        "opencv-python",
        "numba",
        "rdkit",
        "ray",
    ]


def get_extras_require():
    req = {
        "dev": [
            "flake8",
            "pytest",
            "pytest-cov",
            "mypy",
        ],
    }
    return req


setup(
    name='SiPMai',
    version=get_version(),
    description='A Simple Yet Effective Scanning Tunnel Microscope Image Simulator',
    author='Zhiyao Luo, Yaotian Yang, Jiali Li',
    author_email='zhiyao.luo@eng.ox.ac.uk',
    url='http://github.com/GilesLuo/SiPMai',
    packages=find_packages(
        exclude=["test", "test.*", "examples", "examples.*", "docs", "docs.*"]
    ),
    entry_points={
        'console_scripts': ['generate_pubchem=SiPMai.gen_data.gen_all_data_pipeline:main'],
        # 'console_scripts': ['test2=SiPMai.test2:main'],
    },
    install_requires=get_install_requires(),
    package_data={'SiPMai': ['smiles/*.json']},
    include_package_data=True,
    long_description=open("README.md", encoding="utf8").read(),
    long_description_content_type="text/markdown",
    license="MIT",
    python_requires='>=3.6, <3.10',
    keywords=["Chemistry Simulation", "STM Image Synthesis"],
    extras_require=get_extras_require(),
    project_urls={
        "Source Code": "https://github.com/my_username/my_package", # todo
    }
)



from unittest import TestCase
from ..gen_data.ray_generation import ray_gen_main
from ..gen_data.prepare_dataset import gen_index_main
import os
from pkg_resources import resource_filename
import json


class TestDataGeneration(TestCase):

    def test_smiles_exist(self):
        # check file exists in SiPMai/smiles/ folder
        smiles_dir = resource_filename('SiPMai', 'smiles')
        self.assertTrue(os.path.exists(smiles_dir))

    def test_ray_gen_main(self):
        def take(n, iterable):
            """Return the first n items of the iterable as a dict."""
            return dict(list(iterable)[:n])

        smiles_json = resource_filename('SiPMai', 'smiles/pubchem_39_200_100k.json')
        with open(smiles_json, 'r') as f:
            mol_dict = json.load(f)

        n = 10
        mol_dict = take(n, mol_dict.items())
        cmd_dir = os.getcwd()
        mol_save_dir = os.path.join(cmd_dir, "test_gen_data")
        ray_gen_main(mol_dict, mol_save_dir, 128, 24, False, False, False, True)

        # check if mol_save_dir exists
        self.assertTrue(os.path.exists(mol_save_dir))

        img_dir = os.path.join(mol_save_dir, "img")
        json_dir = os.path.join(mol_save_dir, "json")
        info_dir = os.path.join(mol_save_dir, "info")
        # check if mol_save_dir contains n images
        self.assertEqual(len([f for f in os.listdir(img_dir) if "_img" in f and "orig_img" not in f]), n)
        self.assertEqual(len([f for f in os.listdir(img_dir) if "orig_img" in f]), n)
        # check if mol_save_dir contains n json files
        self.assertEqual(len([f for f in os.listdir(json_dir) if f.endswith('.json')]), n)
        # check if mol_save_dir contains n npz files
        self.assertEqual(len([f for f in os.listdir(info_dir) if f.endswith('.npz')]), n)
        self.assertEqual(len([f for f in os.listdir(info_dir) if f.endswith('.png') and "_mol_drawing.png" in f]), n)

    def test_gen_index_main(self):
        cmd_dir = os.getcwd()
        mol_save_dir = os.path.join(cmd_dir, "test_gen_data")
        gen_index_main(mol_save_dir, 0.6, 0.2, 0.2, 1)

        # check if mol_save_dir contains train, val, test folders
        self.assertTrue(os.path.exists(os.path.join(mol_save_dir, "train_set_index.json")))
        self.assertTrue(os.path.exists(os.path.join(mol_save_dir, "val_set_index.json")))
        self.assertTrue(os.path.exists(os.path.join(mol_save_dir, "test_set_index.json")))

if __name__ == '__main__':
    TestCase



import json
import pandas as pd
import numpy as np
from torch.utils.data import Dataset, DataLoader, Sampler
from typing import Tuple, List, Iterator
from random import Random
import threading
import os

class MoleculeDataset(Dataset):
    def __init__(self, dataset_json, transform=None, **kwargs):
        with open(dataset_json, "r") as f:
            self.data_index = json.load(f)
        self.data_dir = "/".join(dataset_json.split("/")[:-1])  # assume dataset json is at the root data dir
        self.cid = list(self.data_index.keys())
        self.img_files = self.get_directory(self.data_dir, "img")
        self.info_files = self.get_directory(self.data_dir, "info")
        self.json_files = self.get_directory(self.data_dir, "json")

        self._batch_Molecule = None

    def batch_Molecule(self):
        raise NotImplementedError

    def get_directory(self, root_dir, key):
        if not key in ["img", "json", "info"]:
            raise KeyError
        # img path is the relative path based on the ROOT_DIR, therefore we need root_dir
        return [os.path.join(root_dir, self.data_index[cid][f"{key}_path"]) for cid in self.cid]

    def get_smiles(self, root_dir):
        smiles = []
        for cid in self.cid:
            with open(os.path.join(root_dir, self.data_index[cid]["json_path"]), "r") as f:
                mol_dict = json.load(f)
            smiles.append(mol_dict["smiles"])
        return smiles

    def get_info(self, npz_file, key):
        if not key in ["arr_atom", "arr_bond", 'adj_matrix', 'molecule_points_height']:
            raise KeyError("Wrong key")
        # Loading compressed data and decompressing it
        data = np.load(npz_file)

        if key in ["arr_atom", "arr_bond", "adj_matrix"]:
            if key == "arr_atom":
                k, s = 'arr_atom', 'arr_atom_shape'
            elif key == "arr_bond":
                k, s = 'arr_bond', 'arr_bond_shape'
            elif key == "adj_matrix":
                k, s = 'adj_matrix', 'adj_shape'
            else:
                raise NotImplementedError
            arr, arr_shape = data[k], data[s]
            arr = np.unpackbits(arr)
            arr = arr.reshape(arr_shape)
            return arr
        elif key == "molecule_points_height":
            return data['molecule_points_height']
        else:
            raise KeyError

    def __len__(self):
        return len(self._data)

    def __getitem__(self, idx):

        return self._data[idx]  # a tuple with four elements



def construct_molecule_batch(data: List[Tuple]) -> MoleculeDataset:
    # This is the collate function
    data = MoleculeDataset(data)  # Re-initialize a small MoleculeDataset object with the batch of data
    data.batch_Molecule()  # Forces computation of the _batch_Molecule

    return data  # a MoleculeDataset with only a batch size


class MoleculeSampler(Sampler):
    """A :class:`MMoleculeSampler` samples data from a :class:`MoleculeDataset` for a :class:`MoleculeDataLoader`."""

    def __init__(self,
                 dataset: MoleculeDataset,
                 shuffle: bool = False,
                 seed: int = 0):
        """
        :param shuffle: Whether to shuffle the data.
        :param seed: Random seed. Only needed if :code:`shuffle` is True.
        """
        super(Sampler, self).__init__()

        self.dataset = dataset
        self.shuffle = shuffle

        self._random = Random(seed)

        self.length = len(self.dataset)

    def __iter__(self) -> Iterator[int]:
        """Creates an iterator over indices to sample."""

        indices = list(range(len(self.dataset)))

        if self.shuffle:
            self._random.shuffle(indices)

        return iter(indices)

    def __len__(self) -> int:
        """Returns the number of indices that will be sampled."""
        return self.length


class MoleculeDataLoader(DataLoader):
    """A :class:`MoleculeDataLoader` is a PyTorch :class:`DataLoader` for loading a :class:`MoleculeDataset`."""

    def __init__(self,
                 dataset: MoleculeDataset,
                 batch_size: int = 50,
                 num_workers: int = 8,
                 shuffle: bool = False,
                 seed: int = 0):
        """
        :param dataset: The :class:`MoleculeDataset` containing the imgs,mols and instructions to load.
        :param batch_size: Batch size.
        :param num_workers: Number of workers used to build batches.
        :param shuffle: Whether to shuffle the data.
        :param seed: Random seed. Only needed if shuffle is True.
        """
        self._dataset = dataset
        self._batch_size = batch_size
        self._num_workers = num_workers
        self._shuffle = shuffle
        self._seed = seed
        self._context = None
        self._timeout = 0
        is_main_thread = threading.current_thread() is threading.main_thread()
        if not is_main_thread and self._num_workers > 0:
            self._context = 'forkserver'  # In order to prevent a hanging
            self._timeout = 3600  # Just for sure that the DataLoader won't hang

        self._sampler = MoleculeSampler(
            dataset=self._dataset,
            shuffle=self._shuffle,
            seed=self._seed
        )

        super(MoleculeDataLoader, self).__init__(
            dataset=self._dataset,
            batch_size=self._batch_size,
            sampler=self._sampler,
            num_workers=self._num_workers,
            collate_fn=construct_molecule_batch,
            multiprocessing_context=self._context,
            timeout=self._timeout
        )

    @property
    def iter_size(self) -> int:
        """Returns the number of data points included in each full iteration through the :class:`MoleculeDataLoader`."""
        return len(self._sampler)

    def __iter__(self) -> Iterator[MoleculeDataset]:
        r"""Creates an iterator which returns :class:`MoleculeDataset`\ s"""
        return super(MoleculeDataLoader, self).__iter__()


if __name__ == '__main__':
    # Test the data loader

    dataset = MoleculeDataset("../pubchem_39_200_100k/train_set_index.json")
    # dataloader = MoleculeDataLoader(dataset)
    # print('cool')
    # for batch in dataloader:
    #     mol_imgs, batch_mol_graph, mol_instructions, labels = batch.batch_Molecule()
    #     print(batch_mol_graph[0].a_scope)
    #     print('cool')
